import logging
import socket
from collections.abc import Collection
from typing import Union

from e7awghal import ClockmasterAu200Hal

from quel_ic_config.box_lock import BoxLockError
from quel_ic_config.exstickge_sock_client import FileLockKeeper
from quel_ic_config.quel1_box import Quel1Box

logger = logging.getLogger(__name__)


class QuelClockMasterV1:
    def __init__(self, ipaddr: str, boxes: Collection[Quel1Box], allow_automatic_boot: bool = True):
        self._ipaddr: str = ipaddr
        self.hal: ClockmasterAu200Hal = ClockmasterAu200Hal(ipaddr=self._ipaddr, auth_callback=self._auth_callback)
        self._boxes: set[Quel1Box] = set(boxes)
        self._allow_automatic_reboot = allow_automatic_boot
        self._lock_keeper: Union[FileLockKeeper, None] = FileLockKeeper(target=(self._ipaddr, 16384))
        if not self._lock_keeper.activate():
            raise BoxLockError(f"failed to acquire lock of {self.__class__.__name__}:{self._ipaddr}")

    def __del__(self):
        self.terminate()

    def check_availability(self) -> bool:
        available: bool = False
        try:
            self.get_current_timecounter()
            available = True
        except socket.timeout:
            logger.info(f"clockmaster {self._ipaddr} doesn't respond to a read-counter request")
        return available

    @property
    def has_lock(self) -> bool:
        return (self._lock_keeper is not None) and self._lock_keeper.has_lock

    def _auth_callback(self) -> bool:
        rv = self.has_lock
        for b in self._boxes:
            rv &= b.has_lock
        return rv

    def get_current_timecounter(self) -> int:
        return self.hal.ctrl.read_counter()

    def sync_boxes(self) -> None:
        kick_done = False
        try:
            self.hal.ctrl.kick_sync({box.wss.hal.syncintf for box in self._boxes})
            kick_done = True
        except socket.timeout:
            if self._allow_automatic_reboot:
                logger.warning(
                    f"clockmaster ({self._ipaddr}) doesn't respond to kick, reboot will be conducted before retrying"
                )
            else:
                logger.error(
                    f"clockmaster ({self._ipaddr}) doesn't respond to kick, "
                    "sending reboot command to clockmaster may solve the problem"
                )
                raise

        if not kick_done:
            self.reboot()
            try:
                self.hal.ctrl.kick_sync({box.wss.hal.syncintf for box in self._boxes})
            except socket.timeout:
                logger.error(f"clockmaster ({self._ipaddr}) doesn't respond to kick after reboot")
                raise

    def reboot(self) -> None:
        self.hal.rebooter.reboot()

    def terminate(self):
        if self._lock_keeper and self._lock_keeper.is_alive():
            if self._lock_keeper.deactivate():
                self._lock_keeper = None



import collections.abc
import json
import logging
import time
from abc import abstractmethod
from enum import Enum, IntEnum
from typing import Any, Collection, Dict, Final, List, Mapping, Sequence, Tuple, Union

import numpy as np
from numpy.typing import NDArray
from pydantic import BaseModel, ConfigDict, RootModel, ValidationError, field_validator, model_validator
from typing_extensions import Self

import adi_ad9082_v170 as adi_ad9082
from adi_ad9082_v170 import ChipTemperatures, CmsError, Device, LinkStatus
from quel_ic_config.abstract_ic import AbstractIcMixin
from quel_ic_config.ad9082_nco import AbstractNcoFtw

logger = logging.getLogger(__name__)


def update_mapping_recursive(d: Dict[str, Any], u: Mapping[str, Any]):
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = update_mapping_recursive(d.get(k, {}), v)
        else:
            d[k] = v
    return d


class NcoFtw(AbstractNcoFtw):
    @classmethod
    def from_ftw(cls, ftw: adi_ad9082.NcoFtw) -> Self:
        enable_fraction = ftw.modulus_a != 0
        x = cls._decode_u48_as_s48(ftw.ftw)
        if enable_fraction:
            b = cls._decode_u48_as_s48(ftw.delta_b)
            a = ftw.modulus_a
        else:
            b = 0
            a = 1
        return cls(ftw=x, delta_b=b, modulus_a=a, enable_fraction=enable_fraction)

    def to_ftw(self) -> adi_ad9082.NcoFtw:
        obj = adi_ad9082.NcoFtw()
        obj.ftw = self._encode_s48_as_u48(self.ftw)
        if self.enable_fraction:
            obj.delta_b = self._encode_s48_as_u48(self.delta_b)
            obj.modulus_a = self.modulus_a
        else:
            obj.delta_b = 0
            obj.modulus_a = 0
        return obj


class NoExtraBaseModel(BaseModel):
    model_config = ConfigDict(extra="forbid", frozen=True)


class FrozenRootModel(RootModel):
    model_config = ConfigDict(frozen=True)


class FrozenSequenceRootModel(FrozenRootModel):
    root: Tuple[Any, ...] = ()

    def __len__(self):
        return len(self.root)

    def __getitem__(self, key):
        return self.root[key]

    def __iter__(self):
        return iter(self.root)


class Ad9082JesdParam(NoExtraBaseModel):
    l: int
    f: int
    m: int
    s: int
    hd: int
    k: int
    n: int
    np: int
    cf: int
    cs: int
    did: int
    bid: int
    lid0: int
    subclass: int
    scr: int
    duallink: int
    jesdv: int
    mode_id: int
    mode_c2r_en: int
    mode_s_sel: int

    def as_cpptype(self) -> adi_ad9082.CmsJesdParam:
        d = adi_ad9082.CmsJesdParam()
        d.l = self.l  # noqa: E741
        d.f = self.f
        d.m = self.m
        d.s = self.s
        d.hd = self.hd
        d.k = self.k
        d.n = self.n
        d.np = self.np
        d.cf = self.cf
        d.cs = self.cs
        d.did = self.did
        d.bid = self.bid
        d.lid0 = self.lid0
        d.subclass = self.subclass
        d.scr = self.scr
        d.duallink = self.duallink
        d.jesdv = self.jesdv
        d.mode_id = self.mode_id
        d.mode_c2r_en = self.mode_c2r_en
        d.mode_s_sel = self.mode_s_sel
        return d


class Ad9082SpiPinConfigEnum(str, Enum):
    SPI_SDO = "SDO"
    SPI_SDIO = "SDIO"

    def as_cpptype(self) -> adi_ad9082.CmsSpiSdoConfig:
        return _Ad9082SpiPinConfigEnum_cpptype_map[self]


_Ad9082SpiPinConfigEnum_cpptype_map: Dict[str, adi_ad9082.CmsSpiSdoConfig] = {
    Ad9082SpiPinConfigEnum.SPI_SDO: adi_ad9082.SPI_SDO,
    Ad9082SpiPinConfigEnum.SPI_SDIO: adi_ad9082.SPI_SDIO,
}


class Ad9082SpiMsbConfigEnum(str, Enum):
    SPI_MSB_FIRST = "FIRST"
    SPI_MSB_LAST = "LAST"

    def as_cpptype(self) -> adi_ad9082.CmsSpiMsbConfig:
        return _Ad9082SpiMsbConfigEnum_cpptype_map[self]


_Ad9082SpiMsbConfigEnum_cpptype_map: Dict[str, adi_ad9082.CmsSpiMsbConfig] = {
    Ad9082SpiMsbConfigEnum.SPI_MSB_FIRST: adi_ad9082.SPI_MSB_FIRST,
    Ad9082SpiMsbConfigEnum.SPI_MSB_LAST: adi_ad9082.SPI_MSB_LAST,
}


class Ad9082SpiAddrNextConfigEnum(str, Enum):
    SPI_ADDR_INC = "INC"
    SPI_ADDR_DEC = "DEC"

    def as_cpptype(self) -> adi_ad9082.CmsSpiAddrInc:
        return _Ad9082SpiAddrNextConfigEnum_cpptype_map[self]


_Ad9082SpiAddrNextConfigEnum_cpptype_map: Dict[str, adi_ad9082.CmsSpiAddrInc] = {
    Ad9082SpiAddrNextConfigEnum.SPI_ADDR_INC: adi_ad9082.SPI_ADDR_INC_AUTO,
    Ad9082SpiAddrNextConfigEnum.SPI_ADDR_DEC: adi_ad9082.SPI_ADDR_DEC_AUTO,
}


class Ad9082SpiConfig(NoExtraBaseModel):
    pin: Ad9082SpiPinConfigEnum
    msb: Ad9082SpiMsbConfigEnum
    addr_next: Ad9082SpiAddrNextConfigEnum


class Ad9082ClockConfig(NoExtraBaseModel):
    ref: int
    dac: int
    adc: int


class _Ad9082SerSwingConfigEnum(IntEnum):
    SWING500 = 500
    SWING750 = 750
    SWING850 = 850
    SWING1000 = 1000

    def as_cpptype(self) -> adi_ad9082.SerSwing:
        return _Ad9082SerSwingConfigEnum_cpptype_map[self]


_Ad9082SerSwingConfigEnum_cpptype_map: Dict[int, adi_ad9082.SerSwing] = {
    _Ad9082SerSwingConfigEnum.SWING500: adi_ad9082.SER_SWING_500,
    _Ad9082SerSwingConfigEnum.SWING750: adi_ad9082.SER_SWING_750,
    _Ad9082SerSwingConfigEnum.SWING850: adi_ad9082.SER_SWING_850,
    _Ad9082SerSwingConfigEnum.SWING1000: adi_ad9082.SER_SWING_1000,
}


class _Ad9082SerPreEmpConfigEnum(IntEnum):
    PREEMP0 = 0
    PREEMP3 = 3
    PREEMP6 = 6

    def as_cpptype(self) -> adi_ad9082.SerPreEmp:
        return _Ad9082SerPreEmpConfigEnum_cpptype_map[self]


_Ad9082SerPreEmpConfigEnum_cpptype_map: Dict[int, adi_ad9082.SerPreEmp] = {
    _Ad9082SerPreEmpConfigEnum.PREEMP0: adi_ad9082.SER_PRE_EMP_0DB,
    _Ad9082SerPreEmpConfigEnum.PREEMP3: adi_ad9082.SER_PRE_EMP_3DB,
    _Ad9082SerPreEmpConfigEnum.PREEMP6: adi_ad9082.SER_PRE_EMP_6DB,
}


class _Ad9082SerPostEmpConfigEnum(IntEnum):
    POSTEMP0 = 0
    POSTEMP3 = 3
    POSTEMP6 = 6
    POSTEMP9 = 9
    POSTEMP12 = 12

    def as_cpptype(self) -> adi_ad9082.SerPostEmp:
        return _Ad9082SerPostEmpConfigEnum_cpptype_map[self]


_Ad9082SerPostEmpConfigEnum_cpptype_map: Dict[int, adi_ad9082.SerPostEmp] = {
    _Ad9082SerPostEmpConfigEnum.POSTEMP0: adi_ad9082.SER_POST_EMP_0DB,
    _Ad9082SerPostEmpConfigEnum.POSTEMP3: adi_ad9082.SER_POST_EMP_3DB,
    _Ad9082SerPostEmpConfigEnum.POSTEMP6: adi_ad9082.SER_POST_EMP_6DB,
    _Ad9082SerPostEmpConfigEnum.POSTEMP9: adi_ad9082.SER_POST_EMP_9DB,
    _Ad9082SerPostEmpConfigEnum.POSTEMP12: adi_ad9082.SER_POST_EMP_12DB,
}


class _Ad9082LaneConfigEnum(IntEnum):
    LANE0 = 0
    LANE1 = 1
    LANE2 = 2
    LANE3 = 3
    LANE4 = 4
    LANE5 = 5
    LANE6 = 6
    LANE7 = 7


class _Ad9082LaneMappingConfig(FrozenSequenceRootModel):
    root: Tuple[
        _Ad9082LaneConfigEnum,
        _Ad9082LaneConfigEnum,
        _Ad9082LaneConfigEnum,
        _Ad9082LaneConfigEnum,
        _Ad9082LaneConfigEnum,
        _Ad9082LaneConfigEnum,
        _Ad9082LaneConfigEnum,
        _Ad9082LaneConfigEnum,
    ]

    def as_cpptype(self, d: Union[None, NDArray]) -> NDArray:
        if d is None:
            d = np.zeros(8, np.uint8)
        for i in range(8):
            d[i] = self.root[i]
        return d


class _Ad9082LaneMask(FrozenSequenceRootModel):
    root: Tuple[bool, bool, bool, bool, bool, bool, bool, bool]

    def as_cpptype(self) -> int:
        u = 0
        for i, v in enumerate(self.root):
            u |= int(v) << i
        return u


class _Ad9082LaneCtleFilterConfig(FrozenSequenceRootModel):
    root: Tuple[int, int, int, int, int, int, int, int]

    def as_cpptype(self, d: Union[None, NDArray]) -> NDArray:
        if d is None:
            d = np.zeros(8, np.uint8)
        for i in range(8):
            d[i] = self.root[i]
        return d


class Ad9082DesConfig(NoExtraBaseModel):
    boost: _Ad9082LaneMask
    invert: _Ad9082LaneMask
    ctle_filter: _Ad9082LaneCtleFilterConfig
    lane_mappings: Tuple[_Ad9082LaneMappingConfig, _Ad9082LaneMappingConfig]


class Ad9082SerConfig(NoExtraBaseModel):
    invert: _Ad9082LaneMask
    swing: Tuple[
        _Ad9082SerSwingConfigEnum,
        _Ad9082SerSwingConfigEnum,
        _Ad9082SerSwingConfigEnum,
        _Ad9082SerSwingConfigEnum,
        _Ad9082SerSwingConfigEnum,
        _Ad9082SerSwingConfigEnum,
        _Ad9082SerSwingConfigEnum,
        _Ad9082SerSwingConfigEnum,
    ]
    pre_emp: Tuple[
        _Ad9082SerPreEmpConfigEnum,
        _Ad9082SerPreEmpConfigEnum,
        _Ad9082SerPreEmpConfigEnum,
        _Ad9082SerPreEmpConfigEnum,
        _Ad9082SerPreEmpConfigEnum,
        _Ad9082SerPreEmpConfigEnum,
        _Ad9082SerPreEmpConfigEnum,
        _Ad9082SerPreEmpConfigEnum,
    ]
    post_emp: Tuple[
        _Ad9082SerPostEmpConfigEnum,
        _Ad9082SerPostEmpConfigEnum,
        _Ad9082SerPostEmpConfigEnum,
        _Ad9082SerPostEmpConfigEnum,
        _Ad9082SerPostEmpConfigEnum,
        _Ad9082SerPostEmpConfigEnum,
        _Ad9082SerPostEmpConfigEnum,
        _Ad9082SerPostEmpConfigEnum,
    ]
    lane_mappings: Tuple[_Ad9082LaneMappingConfig, _Ad9082LaneMappingConfig]


class Ad9082SerdesConfig(NoExtraBaseModel):
    ser: Ad9082SerConfig
    des: Ad9082DesConfig


class _Ad9082FducRateConfigEnum(IntEnum):
    FDUC1 = 1
    FDUC2 = 2
    FDUC3 = 3
    FDUC4 = 4
    FDUC6 = 6
    FDUC8 = 8


class _Ad9082CducRateConfigEnum(IntEnum):
    CDUC12 = 12
    CDUC8 = 8
    CDUC6 = 6
    CDUC4 = 4
    CDUC2 = 2
    CDUC1 = 1


class Ad9082ChannelAssignConfig(NoExtraBaseModel):
    dac0: List[_Ad9082LaneConfigEnum]
    dac1: List[_Ad9082LaneConfigEnum]
    dac2: List[_Ad9082LaneConfigEnum]
    dac3: List[_Ad9082LaneConfigEnum]

    @staticmethod
    def _as_cpptype_sub(chl: Sequence[int]) -> int:
        v: int = 0
        for ch in chl:
            v |= 1 << ch
        return v

    def as_cpptype(self) -> List[int]:
        return [
            self._as_cpptype_sub(self.dac0),
            self._as_cpptype_sub(self.dac1),
            self._as_cpptype_sub(self.dac2),
            self._as_cpptype_sub(self.dac3),
        ]

    @staticmethod
    def dac_order(idx: int, v: List[int]) -> List[int]:
        u = sorted(v, reverse=True)
        if v != u:
            logger.warning(f"param.dac.channel_assign[{idx}] is not sorted in reversed order, fix it")
        return u

    @field_validator("dac0", mode="before")
    def dac0_order(cls, v: List[int]) -> List[int]:
        return cls.dac_order(0, v)

    @field_validator("dac1", mode="before")
    def dac1_order(cls, v: List[int]) -> List[int]:
        return cls.dac_order(1, v)

    @field_validator("dac2", mode="before")
    def dac2_order(cls, v: List[int]) -> List[int]:
        return cls.dac_order(2, v)

    @field_validator("dac3", mode="before")
    def dac3_order(cls, v: List[int]) -> List[int]:
        return cls.dac_order(3, v)


class Ad9082InterpolationRateConfig(NoExtraBaseModel):
    channel: _Ad9082FducRateConfigEnum
    main: _Ad9082CducRateConfigEnum


class Ad9082ShiftFreqConfig(NoExtraBaseModel):
    channel: Tuple[int, int, int, int, int, int, int, int]
    main: Tuple[int, int, int, int]


class Ad9082DacConfig(NoExtraBaseModel):
    jesd204: Ad9082JesdParam
    lane_xbar: Tuple[
        _Ad9082LaneConfigEnum,
        _Ad9082LaneConfigEnum,
        _Ad9082LaneConfigEnum,
        _Ad9082LaneConfigEnum,
        _Ad9082LaneConfigEnum,
        _Ad9082LaneConfigEnum,
        _Ad9082LaneConfigEnum,
        _Ad9082LaneConfigEnum,
    ]
    channel_assign: Ad9082ChannelAssignConfig
    interpolation_rate: Ad9082InterpolationRateConfig
    shift_freq: Ad9082ShiftFreqConfig
    fullscale_current: Tuple[int, int, int, int]


class _Ad9082VirtualConverterConfigEnum(IntEnum):
    VC0 = 0
    VC1 = 1
    VC2 = 2
    VC3 = 3
    VC4 = 4
    VC5 = 5
    VC6 = 6
    VC7 = 7
    VC8 = 8
    VC9 = 9
    VC10 = 10
    VC11 = 11
    VC12 = 12
    VC13 = 13
    VC14 = 14
    VC15 = 15


class _Ad9082ConvSel(FrozenSequenceRootModel):
    root: Tuple[
        _Ad9082VirtualConverterConfigEnum,
        _Ad9082VirtualConverterConfigEnum,
        _Ad9082VirtualConverterConfigEnum,
        _Ad9082VirtualConverterConfigEnum,
        _Ad9082VirtualConverterConfigEnum,
        _Ad9082VirtualConverterConfigEnum,
        _Ad9082VirtualConverterConfigEnum,
        _Ad9082VirtualConverterConfigEnum,
        _Ad9082VirtualConverterConfigEnum,
        _Ad9082VirtualConverterConfigEnum,
        _Ad9082VirtualConverterConfigEnum,
        _Ad9082VirtualConverterConfigEnum,
        _Ad9082VirtualConverterConfigEnum,
        _Ad9082VirtualConverterConfigEnum,
        _Ad9082VirtualConverterConfigEnum,
        _Ad9082VirtualConverterConfigEnum,
    ]

    def as_cpptype(self) -> adi_ad9082.JtxConvSel:
        d = adi_ad9082.JtxConvSel()
        d.virtual_converter0_index = self[0]
        d.virtual_converter1_index = self[1]
        d.virtual_converter2_index = self[2]
        d.virtual_converter3_index = self[3]
        d.virtual_converter4_index = self[4]
        d.virtual_converter5_index = self[5]
        d.virtual_converter6_index = self[6]
        d.virtual_converter7_index = self[7]
        d.virtual_converter8_index = self[8]
        d.virtual_converter9_index = self[9]
        d.virtual_convertera_index = self[10]
        d.virtual_converterb_index = self[11]
        d.virtual_converterc_index = self[12]
        d.virtual_converterd_index = self[13]
        d.virtual_convertere_index = self[14]
        d.virtual_converterf_index = self[15]
        return d


class _Ad9082FddcRateConfigEnum(IntEnum):
    FDDC1 = 1
    FDDC2 = 2
    FDDC3 = 3
    FDDC4 = 4
    FDDC6 = 6
    FDDC8 = 8
    FDDC12 = 12
    FDDC16 = 16
    FDDC24 = 24

    def as_cpptype(self) -> adi_ad9082.AdcFineDdcDcm:
        return _Ad9082FddcRateConfigEnum_cpptype_map[self]

    @classmethod
    def from_cpptype(self, r: int) -> int:
        for k, v in _Ad9082CddcRateConfigEnum_cpptype_map.items():
            if v == r:
                return k
        else:
            raise ValueError(f"invalid AdcFineDdcDcm value: {r}")


_Ad9082FddcRateConfigEnum_cpptype_map: Dict[int, adi_ad9082.AdcFineDdcDcm] = {
    _Ad9082FddcRateConfigEnum.FDDC1: adi_ad9082.ADC_FDDC_DCM_1,
    _Ad9082FddcRateConfigEnum.FDDC2: adi_ad9082.ADC_FDDC_DCM_2,
    _Ad9082FddcRateConfigEnum.FDDC3: adi_ad9082.ADC_FDDC_DCM_3,
    _Ad9082FddcRateConfigEnum.FDDC6: adi_ad9082.ADC_FDDC_DCM_4,
    _Ad9082FddcRateConfigEnum.FDDC8: adi_ad9082.ADC_FDDC_DCM_6,
    _Ad9082FddcRateConfigEnum.FDDC12: adi_ad9082.ADC_FDDC_DCM_8,
    _Ad9082FddcRateConfigEnum.FDDC16: adi_ad9082.ADC_FDDC_DCM_12,
    _Ad9082FddcRateConfigEnum.FDDC24: adi_ad9082.ADC_FDDC_DCM_24,
}


class _Ad9082DecimationRateChannelConfig(FrozenSequenceRootModel):
    root: Tuple[
        _Ad9082FddcRateConfigEnum,
        _Ad9082FddcRateConfigEnum,
        _Ad9082FddcRateConfigEnum,
        _Ad9082FddcRateConfigEnum,
        _Ad9082FddcRateConfigEnum,
        _Ad9082FddcRateConfigEnum,
        _Ad9082FddcRateConfigEnum,
        _Ad9082FddcRateConfigEnum,
    ]

    @model_validator(mode="before")
    @classmethod
    def validate_rates(cls, sq: Any) -> Sequence:
        if not isinstance(sq, Sequence):
            raise ValueError("non-sequence value")
        if len(sq) != 8:
            raise ValueError("invalid length")
        if len(set(sq)) > 1:
            raise ValueError("all the channel decimation rates are expected to be identical currently")
        return sq

    def as_cpptype(self) -> List[int]:
        return [int(x.as_cpptype()) for x in self.root]


class _Ad9082CddcRateConfigEnum(IntEnum):
    CDDC1 = 1
    CDDC2 = 2
    CDDC3 = 3
    CDDC4 = 4
    CDDC6 = 6
    CDDC8 = 8
    CDDC9 = 9
    CDDC12 = 12
    CDDC16 = 16
    CDDC18 = 18
    CDDC24 = 24
    CDDC36 = 36

    def as_cpptype(self) -> adi_ad9082.AdcCoarseDdcDcm:
        return _Ad9082CddcRateConfigEnum_cpptype_map[self]

    @classmethod
    def from_cpptype(self, r: int) -> int:
        for k, v in _Ad9082CddcRateConfigEnum_cpptype_map.items():
            if v == r:
                return k
        else:
            raise ValueError(f"invalid AdcCoarseDdcDcm value: {r}")


_Ad9082CddcRateConfigEnum_cpptype_map: Dict[int, adi_ad9082.AdcCoarseDdcDcm] = {
    _Ad9082CddcRateConfigEnum.CDDC1: adi_ad9082.ADC_CDDC_DCM_1,
    _Ad9082CddcRateConfigEnum.CDDC2: adi_ad9082.ADC_CDDC_DCM_2,
    _Ad9082CddcRateConfigEnum.CDDC3: adi_ad9082.ADC_CDDC_DCM_3,
    _Ad9082CddcRateConfigEnum.CDDC4: adi_ad9082.ADC_CDDC_DCM_4,
    _Ad9082CddcRateConfigEnum.CDDC6: adi_ad9082.ADC_CDDC_DCM_6,
    _Ad9082CddcRateConfigEnum.CDDC8: adi_ad9082.ADC_CDDC_DCM_8,
    _Ad9082CddcRateConfigEnum.CDDC9: adi_ad9082.ADC_CDDC_DCM_9,
    _Ad9082CddcRateConfigEnum.CDDC12: adi_ad9082.ADC_CDDC_DCM_12,
    _Ad9082CddcRateConfigEnum.CDDC16: adi_ad9082.ADC_CDDC_DCM_16,
    _Ad9082CddcRateConfigEnum.CDDC18: adi_ad9082.ADC_CDDC_DCM_18,
    _Ad9082CddcRateConfigEnum.CDDC24: adi_ad9082.ADC_CDDC_DCM_24,
    _Ad9082CddcRateConfigEnum.CDDC36: adi_ad9082.ADC_CDDC_DCM_36,
}


class _Ad9082DecimationRateMainConfig(FrozenSequenceRootModel):
    root: Tuple[
        _Ad9082CddcRateConfigEnum, _Ad9082CddcRateConfigEnum, _Ad9082CddcRateConfigEnum, _Ad9082CddcRateConfigEnum
    ]

    @model_validator(mode="before")
    @classmethod
    def validate_rates(cls, sq: Any) -> Sequence:
        if not isinstance(sq, Sequence):
            raise ValueError("non-sequence value")
        if len(sq) != 4:
            raise ValueError("invalid length")
        if len(set(sq)) > 1:
            raise ValueError("all the main decimation rates are expected to be identical currently")
        return sq

    def as_cpptype(self) -> List[int]:
        return [int(x.as_cpptype()) for x in self.root]


class Ad9082DecimationRateConfig(NoExtraBaseModel):
    channel: _Ad9082DecimationRateChannelConfig
    main: _Ad9082DecimationRateMainConfig


class Ad9082ComplexToRealEnableConfig(NoExtraBaseModel):
    channel: Tuple[bool, bool, bool, bool, bool, bool, bool, bool]
    main: Tuple[bool, bool, bool, bool]


class Ad9082AdcConfig(NoExtraBaseModel):
    jesd204: Tuple[Ad9082JesdParam, Ad9082JesdParam]
    lane_xbar: Tuple[
        _Ad9082LaneConfigEnum,
        _Ad9082LaneConfigEnum,
        _Ad9082LaneConfigEnum,
        _Ad9082LaneConfigEnum,
        _Ad9082LaneConfigEnum,
        _Ad9082LaneConfigEnum,
        _Ad9082LaneConfigEnum,
        _Ad9082LaneConfigEnum,
    ]
    converter_mappings: Tuple[_Ad9082ConvSel, _Ad9082ConvSel]
    decimation_rate: Ad9082DecimationRateConfig
    shift_freq: Ad9082ShiftFreqConfig
    c2r_enable: Ad9082ComplexToRealEnableConfig


class Ad9082Config(NoExtraBaseModel):
    spi: Ad9082SpiConfig
    clock: Ad9082ClockConfig
    serdes: Ad9082SerdesConfig
    dac: Ad9082DacConfig
    adc: Ad9082AdcConfig


class Ad9082Mixin(AbstractIcMixin):
    """wrapping adi_ad9081 APIs to make them Pythonic"""

    _WAIT_FOR_STABILIZATION_OF_LINK: float = 10.0  # [s]

    def __init__(self, name: str):
        super().__init__(name)
        self.device: Final[Device] = Device()
        self.device.callback_set(
            self._read_reg_cb, self._write_reg_cb, self._delay_us_cb, self._log_write_cb, self._reset_pin_ctrl_cb
        )
        # Notes: caches
        self._interp_cache: Union[Tuple[int, int], None] = None
        self._fduc_map_cache: Union[Tuple[Tuple[int, ...], ...], None] = None
        self._dcm_cache: Union[Tuple[int, int], None] = None

    def __del__(self):
        self.device.callback_unset()

    def read_reg(self, addr: int) -> int:
        return self.hal_reg_get(addr)

    def write_reg(self, addr: int, data: int) -> None:
        self.hal_reg_set(addr, data)

    def dump_regs(self) -> Dict[int, int]:
        raise NotImplementedError

    def _read_u48(self, addr) -> int:
        ftw_ftw = bytearray(6)
        for i in range(6):
            ftw_ftw[i] = self.read_reg(addr + i)
        return int.from_bytes(ftw_ftw, byteorder="little", signed=False)

    @abstractmethod
    def _read_reg_cb(self, address: int) -> Tuple[bool, int]:
        pass

    @abstractmethod
    def _write_reg_cb(self, address: int, value: int) -> Tuple[bool]:
        pass

    def _delay_us_cb(self, us: int) -> Tuple[bool]:
        logger.debug(f"delay {us}us")
        time.sleep(us * 1e-6)
        return (True,)

    @abstractmethod
    def _reset_pin_ctrl_cb(self, level: int) -> Tuple[bool]:
        pass

    def _log_write_cb(self, level: int, msg: str) -> Tuple[bool]:
        if level & 0x0001:
            logger.warning(msg)  # error-level log will be generated at the higher layer
        elif level & 0x0002:
            logger.warning(msg)
        elif level & 0x0004:
            logger.info(msg)
        else:
            logger.debug(msg)
        return (True,)

    def device_reset(self, reset_method: adi_ad9082.Reset = adi_ad9082.SOFT_RESET_AND_INIT) -> None:
        logger.info(f"reset {self.name} with {reset_method}")
        rc = adi_ad9082.device_reset(self.device, reset_method)
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(CmsError(rc).name)

    def device_init(self) -> None:
        logger.info(f"init {self.name}")
        rc = adi_ad9082.device_init(self.device)
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(CmsError(rc).name)

    def hal_reg_set(self, addr: int, data: int) -> None:
        addrdata = adi_ad9082.RegData(addr, data)
        rc = adi_ad9082.hal_reg_set(self.device, addrdata)
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(CmsError(rc).name)

    def hal_reg_get(self, addr: int) -> int:
        addrdata = adi_ad9082.RegData(addr)
        rc = adi_ad9082.hal_reg_get(self.device, addrdata)
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(CmsError(rc).name)
        return addrdata.data & 0xFF

    def device_chip_id_get(self) -> adi_ad9082.CmsChipId:
        chip_id = adi_ad9082.CmsChipId()
        rc = adi_ad9082.device_chip_id_get(self.device, chip_id)
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(CmsError(rc).name)
        return chip_id

    def device_clk_config_set(self, dac_clk_hz: int, adc_clk_hz: int, dev_ref_clk_hz: int) -> None:
        logger.info(
            f"clock_config {self.name}: dac_clk = {dac_clk_hz}Hz, "
            f"adc_clk = {adc_clk_hz}Hz, ref_clk = {dev_ref_clk_hz}Hz"
        )
        rc = adi_ad9082.device_clk_config_set(self.device, dac_clk_hz, adc_clk_hz, dev_ref_clk_hz)
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(CmsError(rc).name)

    # Notes: it is possible to refer to the contents of param_spi, param_d, and so on without taking it as an argument.
    #        On the other hand, this relative access of structured data increases the robustness of the code against
    #        the possible modifications of the data structure (e.g., version update of ADI library.)
    def _set_spi_settings(self, param_spi: Ad9082SpiConfig) -> None:
        spi_pin = param_spi.pin.as_cpptype()
        spi_msb = param_spi.msb.as_cpptype()
        spi_addr_inc = param_spi.addr_next.as_cpptype()
        logger.debug(f"spi_pin: {spi_pin}, spi_msb: {spi_msb}, spi_addr_inc: {spi_addr_inc}")
        self.device.spi_conf_set(spi_pin, spi_msb, spi_addr_inc)

    def _set_des_settings(self, param_d: Ad9082DesConfig) -> None:
        self.device.serdes_info.des_settings.boost_mask = param_d.boost.as_cpptype()
        self.device.serdes_info.des_settings.invert_mask = param_d.invert.as_cpptype()
        param_d.ctle_filter.as_cpptype(self.device.serdes_info.des_settings.ctle_filter)
        param_d.lane_mappings[0].as_cpptype(self.device.serdes_info.des_settings.lane_mapping[0])
        param_d.lane_mappings[1].as_cpptype(self.device.serdes_info.des_settings.lane_mapping[1])
        logger.debug(f"ctle_filter: {self.device.serdes_info.des_settings.ctle_filter}")

    def _set_ser_settings(self, param_s: Ad9082SerConfig) -> None:
        self.device.serdes_info.ser_settings.invert_mask = param_s.invert.as_cpptype()
        for i in range(8):
            lane_settings = self.device.serdes_info.ser_settings.lane_settings[i]
            lane_settings[adi_ad9082.SWING_SETTING] = param_s.swing[i].as_cpptype()
            lane_settings[adi_ad9082.PRE_EMP_SETTING] = param_s.pre_emp[i].as_cpptype()
            lane_settings[adi_ad9082.POST_EMP_SETTING] = param_s.post_emp[i].as_cpptype()
        param_s.lane_mappings[0].as_cpptype(self.device.serdes_info.ser_settings.lane_mapping[0])
        param_s.lane_mappings[1].as_cpptype(self.device.serdes_info.ser_settings.lane_mapping[1])

    def _set_serdes_settings(self, param_sd: Ad9082SerdesConfig):
        self._set_des_settings(param_sd.des)
        self._set_ser_settings(param_sd.ser)

    def _validate_settings(self, param_in: Union[str, Dict[str, Any], Ad9082Config]) -> Ad9082Config:
        if isinstance(param_in, str):
            param: Ad9082Config = Ad9082Config.model_validate(json.loads(param_in))
        elif isinstance(param_in, dict):
            param = Ad9082Config.model_validate(param_in)
        elif isinstance(param_in, Ad9082Config):
            # Notes: already validated
            param = param_in
        else:
            raise AssertionError
        return param

    def configure(
        self,
        param_in: Union[str, Dict[str, Any], Ad9082Config],
        *,
        hard_reset: bool = False,
        soft_reset: bool = False,
        use_204b: bool = False,
        use_bg_cal: bool = True,
        wait_after_device_init: float = 0.1,
    ) -> None:
        if use_204b:
            raise RuntimeError("workaround for establishing 204C link with 204B procedure was deprecated")

        param: Ad9082Config = self._validate_settings(param_in)

        ver = adi_ad9082.ApiRevision()
        rc = adi_ad9082.device_api_revision_get(self.device, ver)
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError("failed to retrieve api version")
        logger.info(f"adi ad9082 api version is {ver.major}.{ver.minor}.{ver.rc}")

        dev_ref_clk_hz = int(param.clock.ref)
        dac_clk_hz = int(param.clock.dac)
        adc_clk_hz = int(param.clock.adc)
        sysref_clk_hz = 62500

        rc, common_sysref_clk_hz = adi_ad9082.sync_sysref_frequency_set(
            self.device,
            dev_ref_clk_hz,
            dac_clk_hz,
            adc_clk_hz,
            param.dac.interpolation_rate.main,
            param.dac.interpolation_rate.channel,
            param.adc.decimation_rate.main.as_cpptype(),
            param.adc.decimation_rate.channel.as_cpptype(),
            adi_ad9082.LINK_0,  # Notes: wrong value is provided in ADI's implementation. (not used actually)
            param.dac.jesd204.as_cpptype(),
            [param.adc.jesd204[0].as_cpptype(), param.adc.jesd204[1].as_cpptype()],
        )
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError("failed to calculate common sysref frequency")
        logger.info(f"common sysref frequency is {common_sysref_clk_hz}Hz")

        if common_sysref_clk_hz % sysref_clk_hz != 0:
            raise RuntimeError("sysref_frequency is invalid for the current settings")
        logger.info(f"sysref frequency {sysref_clk_hz}Hz is valid")

        self._set_spi_settings(param.spi)  # values are set to the device object.
        if hard_reset:
            self.device_reset(adi_ad9082.HARD_RESET_AND_INIT)
        elif soft_reset:
            self.device_reset(adi_ad9082.SOFT_RESET_AND_INIT)
        else:
            self.device_init()
        time.sleep(wait_after_device_init)

        self.device_clk_config_set(dac_clk_hz, adc_clk_hz, dev_ref_clk_hz)
        self.device.clk_info.sysref_mode = adi_ad9082.SYSREF_CONT

        # Notes: values are set to device and used in the following startup methods.
        self._set_serdes_settings(param.serdes)
        self._startup_tx(param.dac)
        self._startup_rx(param.adc)
        # TODO: refactoring
        if (
            param.dac.jesd204.subclass != param.adc.jesd204[0].subclass
            or param.dac.jesd204.subclass != param.adc.jesd204[1].subclass
        ):
            logger.warning(
                f"mismatched subclass of RX (= {param.dac.jesd204.subclass}), TX0 (= {param.adc.jesd204[0].subclass}), "
                f"TX1 (= {param.adc.jesd204[1].subclass})"
            )
        self._establish_link(param.dac.jesd204.subclass, use_bg_cal)

    def reconnect(self, dev_ref_clk_hz: int) -> None:
        # Notes: the following two commented out lines are not necessary for reconnect().
        #        self._set_spi_settings(self.param.spi)
        #        self._set_serdes_settings(self.param.serdes)

        if dev_ref_clk_hz == 0:
            raise RuntimeError("PLL is not initialized yet")
        if self.get_pll_bypassed():
            dac_clk_hz = dev_ref_clk_hz
        else:
            raise RuntimeError("internal PLL is not supported")
        adc_clk_hz = dev_ref_clk_hz // self.get_adc_clk_div()
        self.device_clk_config_set(dac_clk_hz, adc_clk_hz, dev_ref_clk_hz)
        self.device.dev_info.dev_rev = self.device_chip_id_get().dev_revision
        rc, _ = adi_ad9082.jesd_rx_bit_rate_get(self.device)
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(f"{CmsError(rc).name} during calculation of bit rate of JESD204C link")

    def _startup_tx(self, param_tx: Ad9082DacConfig) -> None:
        logger.info("starting-up DACs")

        # Notes: clear it before the modification of its corresponding registers.
        self._interp_cache = None
        self._fduc_map_cache = None

        shift_freq = param_tx.shift_freq

        rc = adi_ad9082.device_startup_tx(
            self.device,
            param_tx.interpolation_rate.main,
            param_tx.interpolation_rate.channel,
            param_tx.channel_assign.as_cpptype(),
            list(shift_freq.main),
            list(shift_freq.channel),
            param_tx.jesd204.as_cpptype(),
        )
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(CmsError(rc).name)

        rc = adi_ad9082.jesd_rx_lanes_xbar_set(self.device, adi_ad9082.LINK_0, [int(x) for x in param_tx.lane_xbar])
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(f"{CmsError(rc).name} at configuring rx lanes")

        for i in range(4):
            self.set_fullscale_current(1 << i, param_tx.fullscale_current[i])

    def _startup_rx(self, param_rx: Ad9082AdcConfig) -> None:
        logger.info("starting up ADCs")

        shift_freq = param_rx.shift_freq

        rc = adi_ad9082.device_startup_rx(
            self.device,
            int(adi_ad9082.ADC_CDDC_ALL),
            int(adi_ad9082.ADC_FDDC_ALL),
            list(shift_freq.main),
            list(shift_freq.channel),
            param_rx.decimation_rate.main.as_cpptype(),
            param_rx.decimation_rate.channel.as_cpptype(),
            list(param_rx.c2r_enable.main),
            list(param_rx.c2r_enable.channel),
            [x.as_cpptype() for x in param_rx.jesd204],
            [x.as_cpptype() for x in param_rx.converter_mappings],
        )
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(CmsError(rc).name)

        rc = adi_ad9082.jesd_tx_lanes_xbar_set(self.device, adi_ad9082.LINK_0, [int(x) for x in param_rx.lane_xbar])
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(CmsError(rc).name)

    def clear_crc_error(self):
        self.write_reg(0x05BB, 0x00)
        self.write_reg(0x05BB, 0x01)

    def _establish_link(self, jesd_subclass: int, use_bg_cal: bool) -> None:
        rc = adi_ad9082.jesd_sysref_average_set(self.device, 7)  # Notes: 7 means 2^7. 7 is the maximum.
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(f"failure in jesd_sysref_avarage_set() due to {CmsError(rc).name}")

        rc = adi_ad9082.sync_sysref_input_config_set(self.device, adi_ad9082.COUPLING_AC, adi_ad9082.SIGNAL_LVDS, 0, 0)
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(f"failure in sync_sysref_input_config_set() due to {CmsError(rc).name}")

        rc = adi_ad9082.adc_ddc_coarse_sync_enable_set(self.device, adi_ad9082.ADC_CDDC_ALL.value, 1)
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(f"failure in adc_ddc_coarse_sync_enable_set() due to {CmsError(rc).name}")
        rc = adi_ad9082.adc_ddc_coarse_sync_next_set(self.device, adi_ad9082.ADC_CDDC_ALL.value, 1)
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(f"failure in adc_ddc_coarse_sync_next_set() due to {CmsError(rc).name}")
        rc = adi_ad9082.adc_ddc_coarse_trig_nco_reset_enable_set(self.device, adi_ad9082.ADC_CDDC_ALL.value, 0)
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(f"failure in adc_ddc_coarse_nco_reset_enable_set() due to {CmsError(rc).name}")

        rc = adi_ad9082.adc_ddc_fine_sync_enable_set(self.device, adi_ad9082.ADC_FDDC_ALL.value, 1)
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(f"failure in adc_ddc_fine_sync_enable_set() due to {CmsError(rc).name}")
        rc = adi_ad9082.adc_ddc_fine_sync_next_set(self.device, adi_ad9082.ADC_FDDC_ALL.value, 1)
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(f"failure in adc_ddc_fine_sync_next_set() due to {CmsError(rc).name}")
        rc = adi_ad9082.adc_ddc_fine_trig_nco_reset_enable_set(self.device, adi_ad9082.ADC_FDDC_ALL.value, 0)
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(f"failure in adc_ddc_fine_nco_reset_enable_set() due to {CmsError(rc).name}")

        logger.info("oneshot sync'ing")
        rc = adi_ad9082.jesd_oneshot_sync(self.device, adi_ad9082.CmsJesdSubclass(jesd_subclass))
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(f"failure in jesd_oneshot_sync() due to {CmsError(rc).name}")

        rc, phase = adi_ad9082.jesd_sysref_monitor_phase_get(self.device)
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(f"failure in jesd_sysref_monitor_phase_get() due to {CmsError(rc).name}")
        logger.info(f"phase offset between incoming SYSREF and internal LMFC/LEMC: {phase} DAC clock units")

        logger.info("enabling JESD204C tx link")
        rc = adi_ad9082.jesd_tx_link_enable_set(self.device, adi_ad9082.LINK_0, 1)
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(f"failure in jesd_tx_link_enable_set() due to {CmsError(rc).name}")

        logger.info("enabling JESD204C rx link")
        rc = adi_ad9082.jesd_rx_link_enable_set(self.device, adi_ad9082.LINK_0, 1)
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(f"failure in jesd_rx_link_enable_set() due to {CmsError(rc).name}")

        logger.info("calibrating JESD204C rx link")
        if use_bg_cal:
            logger.info("activating background calibration")
        rc = adi_ad9082.jesd_rx_calibrate_204c(
            self.device, 1, self.device.serdes_info.des_settings.boost_mask, 1 if use_bg_cal else 0
        )
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(f"failure in jesd_rx_calibrate_204c due to {CmsError(rc).name}")

        logger.info("re-enabling JESD204C rx link")
        # Notes this sequence is based on ads9_app of v170.
        rc = adi_ad9082.jesd_rx_link_enable_set(self.device, adi_ad9082.LINK_0, 0)
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(f"failure in jesd_rx_link_enable_set() due to {CmsError(rc).name}")

        rc = adi_ad9082.jesd_rx_link_enable_set(self.device, adi_ad9082.LINK_0, 1)
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(f"failure in jesd_rx_link_enable_set() due to {CmsError(rc).name}")

        rc = adi_ad9082.jesd_rx_link_enable_set(self.device, adi_ad9082.LINK_0, 0)
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(f"failure in jesd_rx_link_enable_set() due to {CmsError(rc).name}")

        rc = adi_ad9082.jesd_rx_link_enable_set(self.device, adi_ad9082.LINK_0, 1)
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(f"failure in jesd_rx_link_enable_set() due to {CmsError(rc).name}")

        logger.info(f"waiting for {self._WAIT_FOR_STABILIZATION_OF_LINK} seconds for link stabilization")
        time.sleep(self._WAIT_FOR_STABILIZATION_OF_LINK)

        self.clear_crc_error()

    def get_link_status(self) -> Tuple[LinkStatus, int]:
        rc, link_status, _ = adi_ad9082.jesd_rx_link_status_get(self.device, adi_ad9082.LINK_0)
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(f"failure in jesd_rx_link_status_get() due to {rc}")
        rc, crc_error = adi_ad9082.jesd_rx_204c_crc_irq_status_get(self.device, adi_ad9082.LINK_0)
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(f"failure in jesd_rx_204c_crc_irq_status_get() due to {rc}")
        return link_status, crc_error

    def dump_jesd_status(self) -> None:
        vals: Dict[int, int] = {}
        for addr in range(0x670, 0x678):
            val = self.hal_reg_get(addr)
            vals[addr] = val
        for addr in (0x702, 0x728, 0x0CA):
            val = self.hal_reg_get(addr)
            vals[addr] = val

        for addr, val in vals.items():
            logger.info(f"*({addr:04x}) = {val:02x}")

    def is_bgcal_enabled(self) -> bool:
        return self.hal_reg_get(0x21C1) & 0x08 == 0x08

    def get_crc_error_counts(self) -> List[int]:
        return [self.hal_reg_get(0x0584 + lane_idx) for lane_idx in range(8)]

    def calc_dac_cnco_ftw(self, shift_hz: float, fractional_mode=False) -> NcoFtw:
        obj = NcoFtw.from_frequency(shift_hz, self.device.dev_info.dac_freq_hz)
        if not fractional_mode:
            obj.round()
        return obj

    def calc_dac_fnco_ftw(self, shift_hz: float, fractional_mode=False) -> NcoFtw:
        obj = NcoFtw.from_frequency(shift_hz, self.device.dev_info.dac_freq_hz // self.get_main_interpolation_rate())
        if not fractional_mode:
            obj.round()
        return obj

    def calc_dac_cnco_freq(self, ftw: NcoFtw) -> float:
        return ftw.to_frequency(self.device.dev_info.dac_freq_hz)

    def calc_dac_fnco_freq(self, ftw: NcoFtw) -> float:
        return ftw.to_frequency(self.device.dev_info.dac_freq_hz // self.get_main_interpolation_rate())

    def set_dac_cnco(self, dacs: Collection[int], ftw: NcoFtw) -> None:
        # dacs is actually cducs.
        dac_mask: adi_ad9082.DacSelect = adi_ad9082.DAC_NONE
        for i in dacs:
            if not 0 <= i <= 3:
                raise ValueError("invalid index of dac: {i}")
            dac_mask |= getattr(adi_ad9082, f"DAC_{i:d}")

        if dac_mask != adi_ad9082.DAC_NONE:
            rc = adi_ad9082.dac_duc_nco_enable_set(self.device, int(dac_mask), int(adi_ad9082.DAC_CH_NONE), 1)
            if rc != CmsError.API_CMS_ERROR_OK:
                raise RuntimeError(f"ad9081.dac_duc_nco_enable_set() failed with error code: {rc}")
            rc = adi_ad9082.dac_duc_nco_ftw_set(self.device, int(dac_mask), int(adi_ad9082.DAC_CH_NONE), ftw.to_ftw())
            if rc != CmsError.API_CMS_ERROR_OK:
                raise RuntimeError(f"ad9081.dac_duc_nco_ftw_set() failed with error code: {rc}")

    def get_dac_cnco(self, dac: int) -> NcoFtw:
        # dac is actually cduc
        if not 0 <= dac <= 3:
            raise ValueError(f"invalid index of dac: {dac}")
        dac_mask: adi_ad9082.DacSelect = getattr(adi_ad9082, f"DAC_{dac:d}")
        rc = adi_ad9082.dac_select_set(self.device, int(dac_mask))
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(f"ad9081.dac_duc_nco_enable_set() failed with error code: {rc}")
        ftw = adi_ad9082.NcoFtw()
        enable_fractional: bool = (self.hal_reg_get(0x01C9) & 0x04) == 0x04
        ftw.ftw = self._read_u48(0x1CB)
        if enable_fractional:
            ftw.modulus_a = self._read_u48(0x1D3)
            ftw.delta_b = self._read_u48(0x1D9)
        else:
            ftw.modulus_a = 0
            ftw.delta_b = 0

        try:
            r = NcoFtw.from_ftw(ftw)
        except ValidationError:
            logger.error(f"invalid FTW is detected for DAC CNCO-#{dac}, ignoring it")
            r = NcoFtw.from_frequency(0, self.device.dev_info.dac_freq_hz)
        return r

    def set_dac_fnco(self, channels: Collection[int], ftw: NcoFtw) -> None:
        # channel is actually fduc
        ch_mask: adi_ad9082.DacChannelSelect = adi_ad9082.DAC_CH_NONE
        for i in channels:
            if not 0 <= i <= 7:
                raise ValueError(f"invalid channel: {i}")
            ch_mask |= getattr(adi_ad9082, f"DAC_CH_{i:d}")

        if ch_mask != adi_ad9082.DAC_CH_NONE:
            rc = adi_ad9082.dac_duc_nco_enable_set(self.device, int(adi_ad9082.DAC_NONE), int(ch_mask), 1)
            if rc != CmsError.API_CMS_ERROR_OK:
                raise RuntimeError(f"ad9081.dac_duc_nco_enable_set() failed with error code: {rc}")
            rc = adi_ad9082.dac_duc_nco_ftw_set(self.device, int(adi_ad9082.DAC_NONE), int(ch_mask), ftw.to_ftw())
            if rc != CmsError.API_CMS_ERROR_OK:
                raise RuntimeError(f"ad9081.dac_duc_nco_ftw_set() failed with error code: {rc}")

    def get_dac_fnco(self, channel: int) -> NcoFtw:
        # channel is actually fduc
        if not 0 <= channel <= 7:
            raise ValueError(f"invalid index of channel: {channel}")
        ch_mask: adi_ad9082.DacChannelSelect = getattr(adi_ad9082, f"DAC_CH_{channel:d}")

        rc = adi_ad9082.dac_chan_select_set(self.device, int(ch_mask))
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(f"ad9081.dac_duc_nco_enable_set() failed with error code: {rc}")

        ftw = adi_ad9082.NcoFtw()

        enable_fractional: bool = (self.hal_reg_get(0x01A0) & 0x04) == 0x04
        ftw.ftw = self._read_u48(0x1A2)
        if enable_fractional:
            ftw.modulus_a = self._read_u48(0x1AA)
            ftw.delta_b = self._read_u48(0x1B0)
        else:
            ftw.modulus_a = 0
            ftw.delta_b = 0

        try:
            r = NcoFtw.from_ftw(ftw)
        except ValidationError:
            logger.error(f"invalid FTW is detected for DAC FNCO-#{channel}, ignoring it")
            r = NcoFtw.from_frequency(0, self.device.dev_info.dac_freq_hz)  # Notes: interpolation rate is omitted.
        return r

    def calc_adc_cnco_ftw(self, shift_hz: float, fractional_mode=False) -> NcoFtw:
        obj = NcoFtw.from_frequency(shift_hz, self.device.dev_info.adc_freq_hz)
        if not fractional_mode:
            obj.round()
        return obj

    def calc_adc_fnco_ftw(self, shift_hz: float, fractional_mode=False) -> NcoFtw:
        # TODO: take index of ADC in the case that different decimation rates are used among ADCs (low priority).
        # TODO: double shift value when c2r is enabled (low priority).
        obj = NcoFtw.from_frequency(shift_hz, self.device.dev_info.adc_freq_hz // self.get_main_decimation_rate())
        if not fractional_mode:
            obj.round()
        return obj

    def calc_adc_cnco_freq(self, ftw: NcoFtw) -> float:
        return ftw.to_frequency(self.device.dev_info.adc_freq_hz)

    def calc_adc_fnco_freq(self, ftw: NcoFtw) -> float:
        return ftw.to_frequency(self.device.dev_info.adc_freq_hz // self.get_main_decimation_rate())

    def set_adc_cnco(self, adcs: Collection[int], ftw: NcoFtw) -> None:
        adc_mask: adi_ad9082.AdcCoarseDdcSelect = adi_ad9082.ADC_CDDC_NONE
        for i in adcs:
            if not 0 <= i <= 3:
                raise ValueError(f"invalid index of adc: {i}")
            adc_mask |= getattr(adi_ad9082, f"ADC_CDDC_{i:d}")

        if adc_mask != adi_ad9082.ADC_CDDC_NONE:
            # TODO: clarify how NCO_ZIF and NCO_VIF works.
            rc = adi_ad9082.adc_ddc_coarse_nco_mode_set(
                self.device,
                int(adc_mask),
                adi_ad9082.ADC_NCO_ZIF if ftw.is_zero() else adi_ad9082.ADC_NCO_VIF,
            )
            if rc != CmsError.API_CMS_ERROR_OK:
                raise RuntimeError(f"ad9081.adc_ddc_coarse_nco_mode_set() failed with error code: {rc}")
            rc = adi_ad9082.adc_ddc_coarse_nco_ftw_set(self.device, int(adc_mask), ftw.to_ftw())
            if rc != CmsError.API_CMS_ERROR_OK:
                raise RuntimeError(f"ad9081.adc_ddc_nco_ftw_set() failed with error code: {rc}")

    def get_adc_cnco(self, adc: int) -> NcoFtw:
        # Note: adc is actually cddc
        if not 0 <= adc <= 3:
            raise ValueError(f"invalid index of adc: {adc}")
        adc_mask: adi_ad9082.AdcCoarseDdcSelect = getattr(adi_ad9082, f"ADC_CDDC_{adc:d}")

        ftw = adi_ad9082.NcoFtw()
        rc = adi_ad9082.adc_ddc_coarse_nco_ftw_get(self.device, int(adc_mask), ftw)
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(f"ad9081.adc_ddc_coarse_nco_ftw_get() failed with error code: {rc}")

        try:
            r = NcoFtw.from_ftw(ftw)
        except ValidationError:
            logger.error(f"invalid FTW is detected ADC CNCO-#{adc}, ignoring it")
            r = NcoFtw.from_frequency(0, self.device.dev_info.adc_freq_hz)
        return r

    def set_adc_fnco(self, channels: Collection[int], ftw: NcoFtw) -> None:
        ch_mask: adi_ad9082.AdcFineDdcSelect = adi_ad9082.ADC_FDDC_NONE
        for i in channels:
            if not 0 <= i <= 7:
                raise ValueError(f"invalid index of adc channel: {i}")
            ch_mask |= getattr(adi_ad9082, f"ADC_FDDC_{i:d}")

        if ch_mask != adi_ad9082.ADC_FDDC_NONE:
            # TODO: clarify how NCO_ZIF and NCO_VIF works.
            rc = adi_ad9082.adc_ddc_fine_nco_mode_set(
                self.device,
                int(ch_mask),
                adi_ad9082.ADC_NCO_ZIF if ftw.is_zero() else adi_ad9082.ADC_NCO_VIF,
            )
            if rc != CmsError.API_CMS_ERROR_OK:
                raise RuntimeError(f"ad9081.adc_ddc_nco_mode_set() failed with error code: {rc}")
            rc = adi_ad9082.adc_ddc_fine_nco_ftw_set(self.device, int(ch_mask), ftw.to_ftw())
            if rc != CmsError.API_CMS_ERROR_OK:
                raise RuntimeError(f"ad9081.adc_ddc_fine_nco_set() failed with error code: {rc}")

    def get_adc_fnco(self, channel: int) -> NcoFtw:
        # Note: channel is actually fddc
        if not 0 <= channel <= 7:
            raise ValueError(f"invalid index of adc channel: {channel}")
        ch_mask: adi_ad9082.AdcFineDdcSelect = getattr(adi_ad9082, f"ADC_FDDC_{channel:d}")

        ftw = adi_ad9082.NcoFtw()
        rc = adi_ad9082.adc_ddc_fine_nco_ftw_get(self.device, int(ch_mask), ftw)
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(f"ad9081.adc_ddc_fine_nco_ftw_get() failed with error code: {rc}")

        try:
            r = NcoFtw.from_ftw(ftw)
        except ValidationError:
            logger.error(f"invalid FTW is detected for ADC FNCO-#{channel}, ignoring it")
            r = NcoFtw.from_frequency(0, self.device.dev_info.adc_freq_hz)  # Notes: decimation rate is omitted.
        return r

    def is_equivalent_dac_cnco(self, freq0: float, freq1: float, fractional_mode: bool = False) -> bool:
        ftw0 = self.calc_dac_cnco_ftw(freq0, fractional_mode)
        ftw1 = self.calc_dac_cnco_ftw(freq1, fractional_mode)
        return ftw0 == ftw1

    def is_equivalent_dac_fnco(self, freq0: float, freq1: float, fractional_mode: bool = False) -> bool:
        ftw0 = self.calc_dac_fnco_ftw(freq0, fractional_mode)
        ftw1 = self.calc_dac_fnco_ftw(freq1, fractional_mode)
        return ftw0 == ftw1

    def is_equivalent_adc_cnco(self, freq0: float, freq1: float, fractional_mode: bool = False) -> bool:
        ftw0 = self.calc_adc_cnco_ftw(freq0, fractional_mode)
        ftw1 = self.calc_adc_cnco_ftw(freq1, fractional_mode)
        return ftw0 == ftw1

    def is_equivalent_adc_fnco(self, freq0: float, freq1: float, fractional_mode: bool = False) -> bool:
        ftw0 = self.calc_adc_fnco_ftw(freq0, fractional_mode)
        ftw1 = self.calc_adc_fnco_ftw(freq1, fractional_mode)
        return ftw0 == ftw1

    def set_fullscale_current(self, dacs: int, current: int, rerun_calibration: bool = True) -> None:
        if dacs == 0 or (dacs & 0x0F) != dacs:
            raise ValueError(f"wrong specifier of DACs {dacs:x}")

        # Notes: this implementation is kept for the historical reason.
        #        40527uA is the historically used fsc, it is called as 40520 by mistake for a while...
        if not ((7000 <= current <= 40000) or (current in {40520, 40527})):
            raise ValueError(f"invalid current {current}uA")

        if current in (40520, 40527):
            # Notes: adi_ad9081_dac_fsc_set() API pose the maximum limit of fsc as 40000uA.
            logger.info("setting fullscale current to 40527uA, that is the conventional value for QuEL-1")
            self.hal_reg_set(0x001B, dacs)
            self.hal_reg_set(0x0117, 0xA0)
            self.hal_reg_set(0x0118, 0xFF)
        else:
            rc = adi_ad9082.dac_fsc_set(self.device, dacs, current, 1 if rerun_calibration else 0)
            if rc != CmsError.API_CMS_ERROR_OK:
                raise RuntimeError(f"{CmsError(rc).name}")

    def decode_fullscale_current(self, r117: int, r118: int) -> int:
        fscmin = (r117 >> 4) & 0x0F
        fscctrl_1_0 = r117 & 0x03
        fscctrl_9_2 = r118 & 0xFF
        return round((fscmin / 16 + ((fscctrl_9_2 << 2) + fscctrl_1_0) / 1024) * 25000)

    def get_fullscale_current(self, dac: int) -> int:
        if not 0 <= dac <= 3:
            raise ValueError(f"invalid index of dac: {dac}")
        dac_mask: adi_ad9082.DacSelect = getattr(adi_ad9082, f"DAC_{dac:d}")
        rc = adi_ad9082.dac_select_set(self.device, int(dac_mask))
        if rc != CmsError.API_CMS_ERROR_OK:
            raise RuntimeError(f"ad9081.dac_duc_nco_enable_set() failed with error code: {rc}")

        r117 = self.hal_reg_get(0x117)
        r118 = self.hal_reg_get(0x118)
        return self.decode_fullscale_current(r117, r118)

    def is_equal_fullscale_current(self, cur0: int, cur1: int):
        # Notes: rounding error is at most 13 (= ceil(25000 / 1024 / 2)).
        # Notes: one of the given values must be calculated from the actual register values.
        return abs(cur0 - cur1) <= 13

    def _dac_xbar_get(self) -> tuple[tuple[int, ...], ...]:
        r: list[tuple[int, ...]] = []
        _, chnl_intrp = self.get_dac_interpolation_rates()
        for i in range(4):
            adi_ad9082.dac_select_set(self.device, (1 << i))
            if chnl_intrp > 1:
                channels_list: list[int] = []
                channels = self.hal_reg_get(0x01BA)
                for j in reversed(range(8)):  # Notes: the order is important!
                    if channels & (0x01 << j) != 0:
                        channels_list.append(j)
                r.append(tuple(channels_list))
            else:
                raise NotImplementedError("channel interpolation rate is assumed to be more than 1")
        return tuple(r)

    def get_fduc_of_dac(self, dac: int) -> Tuple[int, ...]:
        if not 0 <= dac <= 3:
            raise ValueError(f"invalid index of dac: {dac}")

        if self._fduc_map_cache is None:
            self._fduc_map_cache = self._dac_xbar_get()
        return self._fduc_map_cache[dac]

    def get_virtual_adc_select(self) -> List[int]:
        # Notes: 16 comes from the value of JESD M parameter. (see p.68 of UG-1578 rev.A)
        convsel = []
        for i in range(16):
            v = self.read_reg(0x600 + i)
            if v & 0x80:
                convsel.append(-1)
            else:
                convsel.append(v)
        return convsel

    def get_dac_interpolation_rates(self) -> tuple[int, int]:
        # Notes: (main, channel)
        if self._interp_cache is None:
            intrp_mode = self.hal_reg_get(0x01FF)
            self._interp_cache = (int((intrp_mode >> 4) & 0x0F), int(intrp_mode & 0x0F))
        return self._interp_cache

    def get_main_interpolation_rate(self) -> int:
        return self.get_dac_interpolation_rates()[0]

    def get_channel_interpolation_rate(self) -> int:
        return self.get_dac_interpolation_rates()[1]

    def get_adc_decimation_rates(self) -> tuple[int, int]:
        if self._dcm_cache is None:
            cdcms = []
            for i in range(4):
                adi_ad9082.adc_ddc_coarse_select_set(self.device, 1 << i)
                cdcms.append(int(_Ad9082CddcRateConfigEnum.from_cpptype(self.read_reg(0x0282) & 0x0F)))
            adi_ad9082.adc_ddc_coarse_select_set(self.device, 0)
            if len(set(cdcms)) != 1:
                raise RuntimeError(f"unsupported main decimation settings: {cdcms}")

            fdcms = []
            for i in range(8):
                adi_ad9082.adc_ddc_fine_select_set(self.device, 1 << i)
                fdcms.append(int(_Ad9082FddcRateConfigEnum.from_cpptype(self.read_reg(0x0283) & 0x07)))
            adi_ad9082.adc_ddc_fine_select_set(self.device, 0)
            if len(set(fdcms)) != 1:
                raise RuntimeError(f"unsupported channel decimation settings: {fdcms}")

            self._dcm_cache = cdcms[0], fdcms[0]

        return self._dcm_cache

    def get_main_decimation_rate(self) -> int:
        return self.get_adc_decimation_rates()[0]

    def get_channel_decimation_rate(self) -> int:
        return self.get_adc_decimation_rates()[1]

    def get_pll_bypassed(self) -> bool:
        return (self.read_reg(0x0094) & 0x01) == 0x01

    def get_adc_clk_div(self) -> int:
        return (self.read_reg(0x0180) & 0x03) + 1

    def get_temperatures(self) -> Tuple[int, int]:
        temperatures = ChipTemperatures()
        adi_ad9082.device_get_temperature(self.device, temperatures)
        return temperatures.temp_max, temperatures.temp_min



import logging
from dataclasses import dataclass, field
from typing import Dict, Set, Union

from quel_ic_config.abstract_ic import (
    AbstractIcConfigHelper,
    AbstractIcMixin,
    AbstractIcReg,
    b_1bf_bool,
    b_nbf,
    p_1bf_bool,
    p_nbf,
)

logger = logging.getLogger(__name__)


@dataclass
class Ad5328ConfigReg(AbstractIcReg):
    gain_e_h: bool = field(default=False)  # [5]
    gain_a_d: bool = field(default=False)  # [4]
    buf_e_h: bool = field(default=False)  # [3]
    buf_a_d: bool = field(default=False)  # [2]
    vdd_e_h: bool = field(default=False)  # [1]
    vdd_a_d: bool = field(default=False)  # [0]

    def parse(self, v: int) -> None:
        self.gain_e_h = p_1bf_bool(v, 5)
        self.gain_a_d = p_1bf_bool(v, 4)
        self.buf_e_h = p_1bf_bool(v, 3)
        self.buf_a_d = p_1bf_bool(v, 2)
        self.vdd_e_h = p_1bf_bool(v, 1)
        self.vdd_a_d = p_1bf_bool(v, 0)

    def build(self) -> int:
        return (
            b_1bf_bool(self.gain_e_h, 5)
            | b_1bf_bool(self.gain_a_d, 4)
            | b_1bf_bool(self.buf_e_h, 3)
            | b_1bf_bool(self.buf_a_d, 2)
            | b_1bf_bool(self.vdd_e_h, 1)
            | b_1bf_bool(self.vdd_a_d, 0)
        )


@dataclass
class Ad5328DacX(AbstractIcReg):
    data: int = field(default=0)  # [11:0]

    def parse(self, v: int) -> None:
        self.data = p_nbf(v, 11, 0)

    def build(self) -> int:
        return b_nbf(self.data, 11, 0)


@dataclass
class Ad5328PowerDownReg(AbstractIcReg):
    h: bool = field(default=False)  # [7]
    g: bool = field(default=False)  # [6]
    f: bool = field(default=False)  # [5]
    e: bool = field(default=False)  # [4]
    d: bool = field(default=False)  # [3]
    c: bool = field(default=False)  # [2]
    b: bool = field(default=False)  # [1]
    a: bool = field(default=False)  # [0]

    def parse(self, v: int) -> None:
        self.h = p_1bf_bool(v, 7)
        self.g = p_1bf_bool(v, 6)
        self.f = p_1bf_bool(v, 5)
        self.e = p_1bf_bool(v, 4)
        self.d = p_1bf_bool(v, 3)
        self.c = p_1bf_bool(v, 2)
        self.b = p_1bf_bool(v, 1)
        self.a = p_1bf_bool(v, 0)

    def build(self) -> int:
        return (
            b_1bf_bool(self.h, 7)
            | b_1bf_bool(self.g, 6)
            | b_1bf_bool(self.f, 5)
            | b_1bf_bool(self.e, 4)
            | b_1bf_bool(self.d, 3)
            | b_1bf_bool(self.c, 2)
            | b_1bf_bool(self.b, 1)
            | b_1bf_bool(self.a, 0)
        )


Ad5328Regs: Dict[int, type] = {
    0: Ad5328DacX,
    1: Ad5328DacX,
    2: Ad5328DacX,
    3: Ad5328DacX,
    4: Ad5328DacX,
    5: Ad5328DacX,
    6: Ad5328DacX,
    7: Ad5328DacX,
    8: Ad5328ConfigReg,
    12: Ad5328PowerDownReg,
}


Ad5328RegNames: Dict[str, int] = {
    "DacA": 0,
    "DacB": 1,
    "DacC": 2,
    "DacD": 3,
    "DacE": 4,
    "DacF": 5,
    "DacG": 6,
    "DacH": 7,
    "Config": 8,
    "PowerDown": 12,
}


class Ad5328Mixin(AbstractIcMixin):
    Regs = Ad5328Regs
    RegNames = Ad5328RegNames
    _UPDATE = 10
    _RESET = 15  # both data and control are reset

    def __init__(self, name: str):
        super().__init__(name)
        self._carbon_copy: Dict[int, int] = {}
        self._carbon_copy_not_updated: Dict[int, int] = {}

    def dump_regs(self) -> Dict[int, int]:
        # Notes: AD5328 doesn't support register read.
        raise NotImplementedError

    def soft_reset(self) -> None:
        """commiting soft reset.

        :return: None
        """
        for i in range(8):
            self._carbon_copy[i] = 0
        self.write_reg(self._RESET, 0x000)

    @staticmethod
    def _validate_channel(channel: int) -> None:
        if not (0 <= channel <= 7):
            raise ValueError(f"invalid channel: {channel}")

    def update_dac(self) -> None:
        """update the output of the DACs once.
        :return: None
        """
        self._carbon_copy.update(self._carbon_copy_not_updated)
        self.write_reg(self._UPDATE, 0x002)

    def set_output(self, channel: int, value: int, update_dac=False) -> None:
        """setting the output value of a DAC channel.

        :param channel: an index of a DAC channel to be set. The output voltage of the DAC will be updated when
                        update_dac() is called.
        :param value: an output value of the DAC
        :param update_dac: conduct update_dac() if True. Its default value is False.
        :return: None
        """
        self._validate_channel(channel)
        if not (0 <= value <= 4095):
            raise ValueError(f"invalid output value: {value} for channel {channel} of {self.name}")
        self.write_reg(channel, value)
        if update_dac:
            self.update_dac()

    def get_output_carboncopy(self, channel: int) -> Union[int, None]:
        self._validate_channel(channel)
        if channel in self._carbon_copy:
            return self._carbon_copy[channel]
        else:
            return None


class Ad5328ConfigHelper(AbstractIcConfigHelper):
    """Helper class for programming AD5328 with convenient notations. It also work as a storage of register values since
    AD5328 doesn't support register read.
    """

    def __init__(self, ic: Ad5328Mixin):
        super().__init__(ic, no_read=True)
        self.modified: Set[int] = set()

    def _write_reg(self, addr: int, data: Union[int, AbstractIcReg]) -> None:
        super()._write_reg(addr, data)
        self.modified.add(addr)

    def discard(self):
        super().discard()
        self.modified.clear()

    def flush(self, discard_after_flush=False):
        for addr in (8, 12, 0, 1, 2, 3, 4, 5, 6, 7):
            if addr in self.updated and addr in self.modified:
                self.ic.write_reg(addr, self.updated[addr])
        if discard_after_flush:
            self.discard()
        else:
            self.modified.clear()



import asyncio
import atexit
import logging
import os
import queue
import threading
import time
from abc import ABCMeta, abstractmethod
from enum import Enum
from itertools import cycle
from pathlib import Path
from typing import Any, Callable, Final, Mapping, Set, Tuple, Union
from weakref import WeakSet

import aiocoap
import aiocoap.error
import flufl.lock as fl
import ulid
from packaging.version import Version

from quel_ic_config.box_lock import BoxLockError
from quel_ic_config.exstickge_proxy import LsiKindId, _ExstickgeProxyBase
from quel_ic_config.quel_config_common import _DEFAULT_LOCK_DIRECTORY

_DEFAULT_COAP_SERVER_DETECTION_TIMEOUT: Final[float] = 30.0

logger = logging.getLogger(__name__)


class SyncAsyncThunk:
    def __init__(self, data: Any):
        self._data = data
        self._cv = threading.Condition()
        self._result_is_available: bool = False
        self._result: Any = None
        self._exception: Union[Exception, None] = None

    def set_result(self, result: Any) -> None:
        if self._result_is_available:
            raise RuntimeError("result was already put")
        with self._cv:
            self._result = result
            self._result_is_available = True
            self._cv.notify()

    def set_exception(self, exception: Exception) -> None:
        if self._result_is_available:
            raise RuntimeError("result was already put")
        with self._cv:
            self._exception = exception
            self._result_is_available = True
            self._cv.notify()

    def result(self, timeout=None) -> Any:
        with self._cv:
            self._cv.wait_for(lambda: self._result_is_available, timeout=timeout)

        if self._exception is not None:
            raise self._exception
        else:
            return self._result


class _BoxFailedUnlockError(Exception):
    pass


class AbstractSyncAsyncCoapClient(threading.Thread, metaclass=ABCMeta):
    DEFAULT_COAP_PORT = 5683
    DEFAULT_COAP_RESPONSE_TIMEOUT: Final[float] = 10.0

    _DEFAULT_LOOPING_TIMEOUT: Final[float] = 0.25

    _clients: WeakSet["AbstractSyncAsyncCoapClient"] = WeakSet()
    _create_lock: threading.Lock = threading.Lock()

    class _ProximityTransportTuning(aiocoap.numbers.constants.TransportTuning):
        # Notes: this tuning parameter is tailored for the default timeout (= 3.0 second).
        #        i.e., 0.3 * (1+2+4+8) * 1.25 = 5.625 < 10.0
        ACK_TIMEOUT = 0.3
        ACK_RANDOM_FACTOR = 1.25
        MAX_RETRANSMIT = 3

    @classmethod
    def release_lock_all(cls):
        with cls._create_lock:
            for c in cls._clients:
                if c._locked:
                    c._cleanup()
            # Notes: reconsider the following check is effective or not.
            for c in cls._clients:
                for _ in range(10):
                    if not c.has_lock:
                        break
                    time.sleep(cls._DEFAULT_LOOPING_TIMEOUT)
                else:
                    logger.error(f"failed to unlock {c._target[0]} due to timeout")

    @classmethod
    def release_lock_all_at_exit(cls):
        with cls._create_lock:
            for c in cls._clients:
                if c._locked:
                    c._cleanup_at_exit()

    def __init__(self, target: Tuple[str, int], looping_timeout: float = _DEFAULT_LOOPING_TIMEOUT):
        super().__init__()
        self._target = target
        self._request_queue: queue.SimpleQueue[Union[SyncAsyncThunk, None]] = queue.SimpleQueue()
        self._looping_timeout: float = looping_timeout
        self._lock_timestamp: float = 0.0
        self._request_to_init_context: bool = False
        self._locked: bool = False
        self._old_recovery_key: Union[ulid.ULID, None] = None

    def _register_self(self):
        with self._create_lock:
            for c in self._clients:
                if c._target[0] == self._target[0] and c.has_lock:
                    # Notes: should not happen
                    raise RuntimeError(f"duplicated proxy object for {self._target[0]}")
            else:
                self._clients.add(self)

    def _unregister_self(self):
        with self._create_lock:
            if not self.has_lock:
                try:
                    self._clients.remove(self)
                except KeyError:
                    # Notes: this means already removed automatically
                    pass
            else:
                raise _BoxFailedUnlockError(f"try to delete live lock for {self._target[0]}")

    @property
    def has_lock(self) -> bool:
        return self._locked

    def run(self):
        loop = asyncio.new_event_loop()
        task = loop.create_task(
            self._async_main(),
        )

        while not task.done():
            try:
                loop.run_until_complete(task)
            except KeyboardInterrupt:
                raise
            except RuntimeError as e:
                if len(e.args) > 0 and e.args[0] == "cannot schedule new futures after shutdown":
                    # asyncio/base_events.py(647) raises those after the object is deleted.
                    break
                elif len(e.args) > 0 and e.args[0] == "cannot schedule new futures after interpreter shutdown":
                    # concurrent/features/thread.py(169) raises those after the object is deleted.
                    break
                else:
                    raise
            except SystemExit:
                # TODO: understand it carefully.
                #         asyncio/tasks.py(242) raises those after setting them as results,
                #         but we particularly want them back in the loop
                continue

        loop.close()

    def request(self, **args) -> SyncAsyncThunk:
        req = SyncAsyncThunk(args)
        self._request_queue.put(req)
        return req

    def request_and_wait(self, *, timeout: Union[float, None] = None, **args):
        for _ in range(3):
            req = SyncAsyncThunk(args)
            self._request_queue.put(req)
            try:
                return req.result(timeout=timeout)
            except aiocoap.error.NetworkError as e:
                logger.info(f"network error is detected: {e}")
                if not (len(e.args) > 0 and e.args[0].startswith("[Errno 111]")):
                    raise
        else:
            raise aiocoap.error.NetworkError("connection is refused by server repeatedly")

    def terminate(self) -> None:
        if hasattr(self, "_request_queue"):
            self._request_queue.put(None)
            for _ in range(5):
                if not self._locked:
                    break
                time.sleep(self._looping_timeout / 2)
            self._unregister_self()  # Notes: raise RuntimeException for ClientWithDeviceLock holding a lock at exit.

    @abstractmethod
    async def _take_lock(self, context, with_token: bool) -> None: ...

    @abstractmethod
    async def _keep_lock(self, context) -> bool: ...

    @abstractmethod
    async def _release_lock(self, context, key: Union[ulid.ULID, None] = None) -> bool: ...

    @abstractmethod
    def _check_lock_at_host(self, data: Any) -> bool: ...

    async def _async_main(self):
        context = await aiocoap.Context.create_client_context()

        is_first_loop = True
        while True:
            try:
                if self._old_recovery_key is not None:
                    _ = await self._release_lock(context, self._old_recovery_key)
                    self._old_recovery_key = None
                    logger.info(f"the previous lock of {self._target[0]} is released successfully")

                if self._request_to_init_context:
                    _ = await context.shutdown()
                    logger.info(f"recreating communication context of {self._target[0]} for recovery")
                    context = await aiocoap.Context.create_client_context()
                    self._request_to_init_context = False

                if is_first_loop or self.has_lock:
                    if not (await self._keep_lock(context)):
                        logger.info(f"fails to keep lock of {self._target[0]}")
                        await self._release_lock(context)

                req = await asyncio.get_event_loop().run_in_executor(
                    None, lambda: self._request_queue.get(timeout=self._looping_timeout)
                )  # Notes: queue.Empty exception is raised if no requests come.

                if req is None:
                    break
                else:
                    if self._check_lock_at_host(req._data):
                        _ = asyncio.create_task(self._single_access(req, context))
                    else:
                        _ = asyncio.create_task(self._failed_access(req))

            except queue.Empty:
                # Notes: timeout is set to avoid blocking
                pass
            finally:
                is_first_loop = False

        # Notes: executed only when Box object is explicitly deleted.
        logger.info(f"quitting main loop of {self.__class__.__name__} of {self._target[0]}")
        _ = await self._release_lock(context)
        _ = await context.shutdown()

    async def _single_access(self, req: SyncAsyncThunk, context: aiocoap.Context):
        if "transport_tuning" not in req._data:
            msg = aiocoap.Message(**req._data, transport_tuning=self._ProximityTransportTuning)
        else:
            msg = aiocoap.Message(**req._data)
        try:
            res = await context.request(msg).response
            req.set_result(res)
        except aiocoap.error.NetworkError as e:
            if len(e.args) > 0 and e.args[0].startswith("[Errno 111]"):
                self._request_to_init_context = True
            if len(e.args) > 0 and e.args[0] == "Retransmissions exceeded":
                # Notes: QuEL-1 boxes are always connected to the same hub that the Host PC is connected to.
                #        so, both timeout and retransmission exceeded mean the unavailability of the box.
                req.set_result(None)
            else:
                req.set_exception(e)
        except Exception as e:
            req.set_exception(e)

    async def _failed_access(self, req: SyncAsyncThunk):
        req.set_exception(BoxLockError(f"device lock of {self._target[0]} is not available now"))

    @abstractmethod
    def _cleanup(self): ...

    @abstractmethod
    def _cleanup_at_exit(self): ...


# Notes: for bootstrap of CSS and debug. (no lock is required for the bootstrap, actually)
# Notes: be aware that this lock should not be delegated to the access control of e7awghw APIs.
class SyncAsyncCoapClientWithDummyLock(AbstractSyncAsyncCoapClient):
    async def _take_lock(self, context, with_token: bool) -> None:
        self._locked = True  # Notes: not locked, actually.
        # Notes: do not invoke self._register_self() here, it is meaningless and harmful.

    async def _keep_lock(self, context) -> bool:
        if not self._locked:
            await self._take_lock(context, False)
        return True

    async def _release_lock(self, context, key: Union[ulid.ULID, None] = None) -> bool:
        self._release_lock_body()
        return True

    def _release_lock_body(self) -> None:
        self._locked = False

    def _check_lock_at_host(self, data: Any) -> bool:
        return True

    def _cleanup(self):
        self._release_lock_body()

    def _cleanup_at_exit(self):
        pass

    def terminate(self) -> None:
        super().terminate()
        self._release_lock_body()


class SyncAsyncCoapClientWithFileLock(AbstractSyncAsyncCoapClient):
    # Notes: no lock mechanism is unavailable in v1.2.1 and former firmware.
    #        the substitute lock mechanism is implemented.

    _DEFAULT_LOCK_EXPIRATION: Final[float] = 15.0  # [s]
    _FLOCK_TIMEOUT: int = 5  # [s]
    _KILL_CHECK_PERIOD: int = 10  # cycles

    def __init__(
        self,
        target: Tuple[str, int],
        lock_directory: Path = _DEFAULT_LOCK_DIRECTORY,
        looping_timeout: float = AbstractSyncAsyncCoapClient._DEFAULT_LOOPING_TIMEOUT,
    ):
        super().__init__(target=target, looping_timeout=looping_timeout)
        if not lock_directory.is_dir():
            raise RuntimeError(f"lock directory '{lock_directory}' is unavailable")
        self._lockfile: str = str(lock_directory / self._target[0])
        self._lockobj: fl.Lock = fl.Lock(lockfile=self._lockfile, lifetime=int(self._DEFAULT_LOCK_EXPIRATION))
        self._killfile: str = f"{self._lockobj.claimfile}" + ".kill"
        self._kill_check_counter = cycle([True] + (self._KILL_CHECK_PERIOD - 1) * [False])

    async def _take_lock(self, context, with_token: bool) -> None:
        self._lock_timestamp = time.perf_counter()  # no next chance to take a lock.
        try:
            self._lockobj.lock(timeout=self._FLOCK_TIMEOUT)
            self._locked = True
            self._register_self()
            logger.info(f"successfully acquired the lock of {self._target[0]}")
        except fl.TimeOutError:
            logger.error(f"failed to acquire lock of {self._target[0]}")

    async def _keep_lock(self, context) -> bool:
        if self._lock_timestamp == 0.0:
            await self._take_lock(context, False)
        elif self.has_lock:
            if next(self._kill_check_counter):
                if self._check_if_killfile_exists():
                    logger.info(f"Cancelled to refresh lock of {self._target[0]} due to the killfile.")
                    return False

            try:
                self._lockobj.refresh()
                self._lock_timestamp = time.perf_counter()
            except fl.NotLockedError:
                logger.error(f"failed to refresh lock of {self._target[0]}")
                return False
        return self.has_lock

    def _check_if_killfile_exists(self) -> bool:
        return os.path.exists(self._killfile)

    async def _release_lock(self, context, key: Union[ulid.ULID, None] = None) -> bool:
        self._release_lock_body()
        return True

    def _release_lock_body(self) -> None:
        self._lockobj.unlock(unconditionally=True)
        self._locked = False

    def _check_lock_at_host(self, data: Any) -> bool:
        return self.has_lock

    def _cleanup(self):
        self._release_lock_body()

    def _cleanup_at_exit(self):
        self._release_lock_body()


class SyncAsyncCoapClientWithDeviceLock(AbstractSyncAsyncCoapClient):
    _LOCK_EXPIRATION = 60.0  # [s]
    _LOCK_REACQUIRE_PERIOD = 45.0  # [s]
    _RECOVERY_KEYLEN = 10  # Notes: this comes from option length constraint of CoAP without enabling
    #        CONFIG_COAP_EXTENDED_OPTIONS_LEN.
    _RECOVERY_KEY_DIRECTORY: Path = Path(os.getenv("HOME", "/tmp")) / ".quelware" / "recovery_keys"

    def __init__(
        self,
        target: Tuple[str, int],
        looping_timeout: float = AbstractSyncAsyncCoapClient._DEFAULT_LOOPING_TIMEOUT,
    ):
        super().__init__(target=target, looping_timeout=looping_timeout)
        self._target: Tuple[str, int] = target
        self._old_recovery_key = self._load_old_recovery_key()
        self._recovery_key: ulid.ULID = ulid.ULID()

    def _load_old_recovery_key(self) -> Union[ulid.ULID, None]:
        filepath = self._RECOVERY_KEY_DIRECTORY / f"{self._target[0]}.txt"
        if os.path.exists(filepath):
            try:
                with open(filepath) as f:

                    k = ulid.ULID.from_str(f.read())
                os.remove(filepath)  # Notes: saved key can be used only once.
                delta = time.time() - k.timestamp
                if delta < self._LOCK_EXPIRATION * 1.2:
                    return k
                else:
                    logger.info(f"the previous key is created {round(delta)} seconds ago, already expired")
            except Exception as e:
                logger.info(f"failed to load the previous key due to {e}")
        else:
            logger.info("no previous key is found")
        return None

    def _shorten_key(self, k: Union[ulid.ULID, None] = None) -> str:
        if k is None:
            return str(self._recovery_key)[-self._RECOVERY_KEYLEN :]
        else:
            return str(k)[-self._RECOVERY_KEYLEN :]

    async def _take_lock(self, context, with_key: bool) -> None:
        self._lock_timestamp = time.perf_counter()
        if with_key:
            msg = aiocoap.Message(
                code=aiocoap.GET, uri=f"coap://{self._target[0]}/lock/acquire?k={self._shorten_key()}"
            )
            res = await context.request(msg).response
            if res.code.is_successful():
                self._locked = True
                self._register_self()
                logger.info(f"successfully acquired the lock of {self._target[0]}")
            elif res.code == aiocoap.FORBIDDEN:
                raise BoxLockError(f"device lock of {self._target[0]} is not available now")
            else:
                raise RuntimeError(f"failed to acquire the lock of {self._target[0]}")
        else:
            msg = aiocoap.Message(code=aiocoap.GET, uri=f"coap://{self._target[0]}/lock/acquire")
            res = await context.request(msg).response
            if not res.code.is_successful():
                self._locked = False
                raise RuntimeError(f"failed to extend the lock of {self._target[0]}")

    async def _keep_lock(self, context) -> bool:
        if self._lock_timestamp == 0.0:
            try:
                await self._take_lock(context, with_key=True)
            except BaseException as e:
                logger.warning(e)
        elif self._locked:
            if self._request_to_init_context:
                # Notes: reacquiring the lock with my token
                await self._take_lock(context, with_key=True)
            else:
                cur = time.perf_counter()
                if cur - self._lock_timestamp >= self._LOCK_REACQUIRE_PERIOD:
                    # Notes: just extending a lock
                    await self._take_lock(context, with_key=False)
        return self._locked

    async def _release_lock(self, context, key: Union[ulid.ULID, None] = None) -> bool:
        if key is None:
            msg = aiocoap.Message(code=aiocoap.GET, uri=f"coap://{self._target[0]}/lock/release")
        else:
            msg = aiocoap.Message(
                code=aiocoap.GET, uri=f"coap://{self._target[0]}/lock/release?k={self._shorten_key(key)}"
            )

        try:
            res = await context.request(msg).response
            if res.code.is_successful():
                self._locked = False
                logger.info(f"lock of {self._target[0]} is released successfully")
                return True
            else:
                # Notes: try to unlock irrespective of has_lock, however, don't show failure log if not has_lock.
                if self.has_lock:
                    logger.warning(f"failed to release the lock of {self._target[0]} with code: {res.code}")
                return False
        except Exception as e:
            logger.warning(f"failed to release the lock of {self._target[0]} with exception: {e}")
            return False

    def _check_lock_at_host(self, data: Any) -> bool:
        # Notes: host check is skipped. firmware is in charge of judging permission.
        return True

    def _cleanup(self):
        if hasattr(self, "_request_queue"):
            res1 = self.request_and_wait(
                code=aiocoap.GET,
                uri=f"coap://{self._target[0]}/lock/release",
                timeout=self.DEFAULT_COAP_RESPONSE_TIMEOUT,
            )
            if res1.code.is_successful():
                self._locked = False
                logger.info(f"lock of {self._target[0]} is released successfully")

    def _cleanup_at_exit(self):
        os.makedirs(self._RECOVERY_KEY_DIRECTORY, exist_ok=True)
        filepath = self._RECOVERY_KEY_DIRECTORY / f"{self._target[0]}.txt"
        with open(filepath, "w") as f:
            # Notes: write the recovery key after updating its timestamp (i know this is an unusual usage of ULID)
            z = ulid.ULID()
            f.write(f"{str(z)[:10]}{str(self._recovery_key)[10:]}")
            f.flush()
            f.close()
        logger.info(f"recovery key of {self._target[0]} is written out to {filepath}")

    def terminate(self) -> None:
        try:
            super().terminate()
        except _BoxFailedUnlockError:
            pass


atexit.register(AbstractSyncAsyncCoapClient.release_lock_all_at_exit)


class Quel1seBoard(str, Enum):
    MIXER0 = "mx0"
    MIXER1 = "mx1"
    POWER = "pwr"
    PATHSEL0 = "ps0"
    PATHSEL1 = "ps1"


class _ExstickgeCoapClientBase(_ExstickgeProxyBase):
    DEFAULT_PORT: int = AbstractSyncAsyncCoapClient.DEFAULT_COAP_PORT
    DEFAULT_RESPONSE_TIMEOUT: float = AbstractSyncAsyncCoapClient.DEFAULT_COAP_RESPONSE_TIMEOUT

    _URI_MAPPINGS: Mapping[Tuple[LsiKindId, int], str]
    _READ_REG_PATHS: Mapping[LsiKindId, Callable[[int], str]]
    _WRITE_REG_PATHS_AND_PAYLOADS: Mapping[LsiKindId, Callable[[int, int], Tuple[str, str]]]

    _AVAILABLE_BOARDS: Tuple[Quel1seBoard, ...]

    _DATA_MASKS: Mapping[LsiKindId, int] = {
        LsiKindId.AD9082: 0xFF,
        LsiKindId.ADRF6780: 0xFFFF,
        LsiKindId.LMX2594: 0xFFFF,
        LsiKindId.AD5328: 0xFFFF,
        LsiKindId.MIXERBOARD_GPIO: 0xFFFF,
        LsiKindId.PATHSELECTORBOARD_GPIO: 0x003F,
        LsiKindId.AD7490: 0xFFFF,
        LsiKindId.POWERBOARD_PWM: 0xFFFF,
    }

    _VALID_BOXTYPE: Set[str] = set()
    _VERSION_SPEC: Tuple[Version, Version, Set[Version]] = Version("0.0.0"), Version("0.0.0"), set()

    @classmethod
    def matches(cls, btstr: str, vstr: str):
        minv, maxv, excvs = cls._VERSION_SPEC
        v = Version(vstr)
        return (btstr in cls._VALID_BOXTYPE) and (minv <= v <= maxv) and (v not in excvs)

    def __init__(
        self,
        target_address: str,
        target_port: int = DEFAULT_PORT,
        timeout: float = DEFAULT_RESPONSE_TIMEOUT,
    ):
        super().__init__(target_address, target_port, timeout)
        self._core: AbstractSyncAsyncCoapClient = self._creating_core()
        self._core.start()

    @property
    def has_lock(self) -> bool:
        return self._core.has_lock

    @property
    def available_boards_with_cpld(self) -> Tuple[Quel1seBoard, ...]:
        return self._AVAILABLE_BOARDS

    def _creating_core(self) -> AbstractSyncAsyncCoapClient:
        return SyncAsyncCoapClientWithDummyLock(target=self._target)

    def _coap_return_check(self, res: Union[aiocoap.Message, None], uri: str, raise_exception: bool = True) -> bool:
        if res is None:
            if raise_exception:
                raise RuntimeError(f"failed access to the end-point '{uri}' due to timeout")
            else:
                logger.error(f"failed access to the end-point '{uri}' due to timeout")
                return False
        if not res.code.is_successful():
            # Notes: DeviceLockException is not subject to be controlled by raise_exception argument.
            if res.code == aiocoap.FORBIDDEN:
                raise BoxLockError(f"device lock of {self._target[0]} is not available now")
            if raise_exception:
                raise RuntimeError(f"failed access to the end-point '{uri}' with an error code {res.code}")
            else:
                logger.error(f"failed access to the end-point '{uri}' with an error code {res.code}")
                return False
        return True

    def read_fpga_version(self) -> Tuple[str, str]:
        uri1 = f"coap://{self._target[0]}/version/fpga"
        uri2 = f"coap://{self._target[0]}/hash/fpga"
        res1 = self._core.request_and_wait(code=aiocoap.GET, uri=uri1, timeout=self._timeout)
        res2 = self._core.request_and_wait(code=aiocoap.GET, uri=uri2, timeout=self._timeout)
        self._coap_return_check(res1, uri1)
        self._coap_return_check(res2, uri2)
        return res1.payload.decode(), res2.payload.decode()

    def read_firmware_version(self) -> str:
        uri1 = f"coap://{self._target[0]}/version/firmware"
        res1 = self._core.request_and_wait(code=aiocoap.GET, uri=uri1, timeout=self._timeout)
        self._coap_return_check(res1, uri1)
        return res1.payload.decode()

    def read_boxtype(self) -> str:
        # Notes: some early-stage firmware doesn't implement this API.
        #        those will be eliminated soon.
        uri1 = f"coap://{self._target[0]}/conf/boxtype"
        res1 = self._core.request_and_wait(code=aiocoap.GET, uri=uri1, timeout=self._timeout)
        self._coap_return_check(res1, uri1)
        return res1.payload.decode()

    def read_current_config(self) -> str:
        # Notes: some early-stage firmware doesn't implement this API.
        #        those will be eliminated soon.
        uri1 = f"coap://{self._target[0]}/conf/current"
        res1 = self._core.request_and_wait(code=aiocoap.GET, uri=uri1, timeout=self._timeout)
        self._coap_return_check(res1, uri1)
        return res1.payload.decode()

    def write_current_config(self, cfg: str):
        # Notes: some early-stage firmware doesn't implement this API.
        #        those will be eliminated soon.
        if len(cfg) > 256:
            raise ValueError("too long config data")
        uri1 = f"coap://{self._target[0]}/conf/current"
        res1 = self._core.request_and_wait(code=aiocoap.PUT, payload=cfg.encode(), uri=uri1, timeout=self._timeout)
        self._coap_return_check(res1, uri1)

    def read_board_active(self, board: Quel1seBoard) -> bool:
        if board in self._AVAILABLE_BOARDS:
            uri1 = f"coap://{self._target[0]}/{board.value}/xbar/reset"
            res1 = self._core.request_and_wait(code=aiocoap.GET, uri=uri1, timeout=self._timeout)
            if res1 is not None:
                if res1.code.is_successful():
                    v = res1.payload.decode()
                    if v == "0":
                        return False  # the board is in reset state
                    elif v == "1":
                        return True
                    else:
                        raise RuntimeError(f"unexpected return value: '{res1}' from the end-point '{uri1}'")
                else:
                    raise RuntimeError(f"failed to read active status of the board {board}")
            else:
                raise RuntimeError(f"failed access to the end-point '{uri1}'")
        else:
            raise ValueError(f"the specified board {board} is not available")

    def write_board_active(self, board: Quel1seBoard, active: bool) -> None:
        if board in self._AVAILABLE_BOARDS:
            uri1 = f"coap://{self._target[0]}/{board.value}/xbar/reset"
            v = "1" if active else "0"
            res1 = self._core.request_and_wait(code=aiocoap.PUT, payload=v.encode(), uri=uri1, timeout=self._timeout)
            self._coap_return_check(res1, uri1)
        else:
            raise ValueError(f"the specified board {board} is not available")

    def read_reset(self, kind: LsiKindId, idx: int) -> Union[int, None]:
        if (kind, idx) not in self._URI_MAPPINGS:
            raise ValueError(f"invalid IC ({kind.name}, {idx})")

        if kind == LsiKindId.AD9082:
            uri = f"coap://{self._target[0]}/{self._URI_MAPPINGS[kind, idx]}/reset"
            res = self._core.request_and_wait(code=aiocoap.GET, uri=uri, timeout=self._timeout)
            self._coap_return_check(res, uri)
            return int(res.payload.decode())
        else:
            return None

    def write_reset(self, kind: LsiKindId, idx: int, value: int) -> bool:
        if (kind, idx) not in self._URI_MAPPINGS:
            raise ValueError(f"invalid IC ({kind.name}, {idx})")
        if value not in {0, 1}:
            raise ValueError(f"invalid value for reset: {value}")

        if kind == LsiKindId.AD9082:
            uri = f"coap://{self._target[0]}/{self._URI_MAPPINGS[kind, idx]}/reset"
            res = self._core.request_and_wait(
                code=aiocoap.PUT, uri=uri, payload=str(value).encode(), timeout=self._timeout
            )
            self._coap_return_check(res, uri)
            return True  # processed
        else:
            return False  # delegated

    def read_reg(self, kind: LsiKindId, idx: int, addr: int) -> Union[int, None]:
        if (kind, idx) not in self._URI_MAPPINGS:
            raise ValueError(f"invalid IC ({kind.name if isinstance(kind, LsiKindId) else kind}, {idx})")
        if (addr & self._ADDR_MASKS[kind]) != addr:
            raise ValueError(f"address {addr} of {kind.name} is out of range ")

        if kind in self._READ_REG_PATHS:
            uri = f"coap://{self._target[0]}/{self._URI_MAPPINGS[kind, idx]}" + self._READ_REG_PATHS[kind](addr)
        else:
            raise ValueError(f"no read register API is available for {kind}")

        res = self._core.request_and_wait(code=aiocoap.GET, uri=uri, timeout=self._timeout)
        if self._coap_return_check(res, uri, False):
            return int(res.payload.decode(), 16) & self._DATA_MASKS[kind]
        else:
            return None

    def write_reg(self, kind: LsiKindId, idx: int, addr: int, value: int) -> bool:
        if (kind, idx) not in self._URI_MAPPINGS:
            raise ValueError(f"invalid IC ({kind.name if isinstance(kind, LsiKindId) else kind}, {idx})")
        if (addr & self._ADDR_MASKS[kind]) != addr:
            raise ValueError(f"address {addr} of {kind.name} is out of range ")
        if (value & self._DATA_MASKS[kind]) != value:
            raise ValueError(f"value {value} of {kind.name} is out of range ")

        if kind in self._WRITE_REG_PATHS_AND_PAYLOADS:
            api_path, payload = self._WRITE_REG_PATHS_AND_PAYLOADS[kind](addr, value)
            uri = f"coap://{self._target[0]}/{self._URI_MAPPINGS[kind, idx]}" + api_path
        else:
            raise ValueError(f"no write register API is available for {kind}")

        res = self._core.request_and_wait(code=aiocoap.PUT, uri=uri, payload=payload.encode(), timeout=self._timeout)
        return self._coap_return_check(res, uri, False)

    def terminate(self):
        if hasattr(self, "_core"):
            self._core.terminate()
            self._core.join()
            if self._core._locked:
                logger.warning(
                    f"you should delete a box object ({self._target[0]}) explicitly to release its lock properly"
                )


def get_exstickge_server_info(
    ipaddr_css: str,
    port: int = AbstractSyncAsyncCoapClient.DEFAULT_COAP_PORT,
    timeout_ping: float = _DEFAULT_COAP_SERVER_DETECTION_TIMEOUT,
    timeout_coap: float = AbstractSyncAsyncCoapClient.DEFAULT_COAP_RESPONSE_TIMEOUT,
) -> Tuple[bool, str, str]:
    # Notes: assuming ipaddr_css is already validated
    # Notes: timeout_ping of 10.0 second comes from the booting duration of the server
    #        it was previously implemented with ping3.
    # Notes: port is not currently used
    uri1 = f"coap://{ipaddr_css}/version/firmware"
    uri2 = f"coap://{ipaddr_css}/conf/boxtype"
    res1: Union[aiocoap.Message, None] = None
    res2: Union[aiocoap.Message, None] = None

    # Notes: reading box information with temporary coap client
    core = SyncAsyncCoapClientWithDummyLock(target=(ipaddr_css, port))
    core.start()

    t0 = time.perf_counter()
    while (time.perf_counter() < t0 + timeout_ping) and (res1 is None):
        try:
            res1 = core.request_and_wait(code=aiocoap.GET, uri=uri1, timeout=timeout_coap)
            if res1 is None:
                logger.info(f"no CoAP response from {ipaddr_css}")
                time.sleep(3)
        except aiocoap.error.NetworkError:
            core.terminate()
            time.sleep(3)
            core = SyncAsyncCoapClientWithDummyLock(target=(ipaddr_css, port))
            core.start()
            res1 = None

    if res1 is not None:
        res2 = core.request_and_wait(code=aiocoap.GET, uri=uri2, timeout=timeout_coap)
    core.terminate()
    del core

    # Notes: validating the box information
    if res1 is None:
        logger.info(f"timed out the detection of CoAP server on {ipaddr_css}")
        return False, "", ""  # UDP server, probably
    elif not res1.code.is_successful():
        raise RuntimeError(f"failed to access to end-point '{uri1}' with error code {res1.code}")
    else:
        v = res1.payload.decode()
        logger.info(f"CoAP server version of {ipaddr_css} is '{v}'")
        if res2 and res2.code.is_successful():
            u = res2.payload.decode()
            logger.info(f"reported boxtype of {ipaddr_css} is '{u}'")
            return True, v, u
        else:
            logger.info(f"boxtype information of {ipaddr_css} is not available due to error")
            return True, v, ""
